/******************************************************************
*******************************************************************
***																***
***		BonFriioView application implementation 				***
***																***
*******************************************************************
******************************************************************/

#include <windows.h>
#include <stdio.h>
#include "lernel32.h"
#include "IBonDriver2.h"

/******************************************************************
***																***
***		Local constants/macros definition   					***
***																***
******************************************************************/

#define PART_OF_TARGET_DEVICE_NAME "\\\\?\\USB#VID_"
#define IOCODE_CONTROL		0x00220020UL	// `[iRg[
#define IOCODE_SETTSSIZE	0x00220038UL	// TSXg[TCYݒ
#define IOCODE_ABORTPIPE	0x00220044UL	// TSXg[~
#define IOCODE_TSDATA		0x0022004BUL	// TSXg[
#define TS_SIZE (188*256)
#define RINGBUF_SIZE 256

/******************************************************************
***																***
***		Local static variables definition   					***
***																***
******************************************************************/

								/*************************************/
								/**	BonDriver related   			**/
								/*************************************/
static BOOL bBonOk = FALSE ;
static HMODULE hDllBon = NULL ;
static IBonDriver2* pInsBon = NULL ;
								/*************************************/
								/**	second thread related   		**/
								/*************************************/
static HANDLE hEventThread = NULL ;
static HANDLE hThread = NULL ;

/******************************************************************
***																***
***		Local function prototype definitions					***
***																***
******************************************************************/

static BOOL OpenBonDriver( void ) ;
static void CloseBonDriver( void ) ;
static void HookChannelSetting(
	LPVOID lpInBuffer, DWORD nInBufferSize ) ;
static void BonSetChannel( DWORD dwSpace, DWORD dwChannel ) ;
static void PurgeTsData( void ) ;
static DWORD WINAPI TsGetThread( LPVOID lpParameter ) ;

/******************************************************************
***																***
***		Name:	Setup for application   						***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

BOOL SetupApp( void )
{
								/*************************************/
								/**	initialize variables			**/
								/*************************************/
	bBonOk = FALSE ;
	hDllBon = NULL ;
	pInsBon = NULL ;
	hEventThread = NULL ;
	hThread = NULL ;
	return TRUE ;
}

/******************************************************************
***																***
***		Name:	Cleanup for application 						***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

void CleanupApp( void )
{
}

/******************************************************************
***																***
***		Name:	Answer if the device is target  				***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

BOOL IsTargetDevice( LPCSTR lpFileName )
{
	if ( strstr( lpFileName, PART_OF_TARGET_DEVICE_NAME ) ) {
		DebugPrintF( "Open : %s\n", lpFileName ) ;
		return TRUE ;
	}
	else return FALSE ;
}

/******************************************************************
***																***
***		Name:	Target is opened								***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

void TargetOpened( HANDLE hDevice )
{
								/*************************************/
								/**	open BonDriver  				**/
								/*************************************/
	bBonOk = OpenBonDriver() ;
}

/******************************************************************
***																***
***		Name:	Target is closed								***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

void TargetClosed( void )
{
								/*************************************/
								/**	close BonDriver 				**/
								/*************************************/
	bBonOk = FALSE ;
	CloseBonDriver() ;
}

/******************************************************************
***																***
***		Name:	Override DeviceIoControl						***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

BOOL OrDeviceIoControl( BOOL& bRet,
	HANDLE& hDevice, DWORD& dwIoControlCode,
	LPVOID& lpInBuffer, DWORD& nInBufferSize,
	LPVOID& lpOutBuffer, DWORD& nOutBufferSize, LPDWORD& lpBytesReturned,
	LPOVERLAPPED& lpOverlapped )
{
								/*************************************/
								/**	if not BonDriver opened,		**/
								/**		fail this function  		**/
								/*************************************/
	if ( !bBonOk ) {
		SetLastError( ERROR_GEN_FAILURE ) ;
		bRet = FALSE ;
		return TRUE ;
	}
								/*************************************/
								/**	Get TS data 					**/
								/*************************************/
	if ( dwIoControlCode == IOCODE_TSDATA ) {
		return FALSE ; // not implement yet
	}
								/*************************************/
								/**	Purge TS DATA					**/
								/*************************************/
	if ( dwIoControlCode == IOCODE_ABORTPIPE ) {
		PurgeTsData() ;
	}
								/*************************************/
								/**	hook channnel setting   		**/
								/*************************************/
	else if ( dwIoControlCode == IOCODE_CONTROL ) {
		HookChannelSetting( lpInBuffer, nInBufferSize ) ;
	}
								/*************************************/
								/**	call original kernel32.dll  	**/
								/*************************************/
	return FALSE ;
}

/******************************************************************
***																***
***		Name:	Override CancelIo   							***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

BOOL OrCancelIo( BOOL& bRet,
	HANDLE& hFile )
{
	DebugPrintF( "CancelIo() is called.\n" ) ;
	return FALSE ;
}

/******************************************************************
***																***
***		Name:	Override GetOverlappedResult					***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

BOOL OrGetOverlappedResult( BOOL& bRet,
	HANDLE& hFile, LPOVERLAPPED& lpOverlapped,
	LPDWORD& lpNumberOfBytesTransferred, BOOL& bWait )
{
	return FALSE ;
}

/******************************************************************
***																***
***		Local function to implement BonFriioView				***
***																***
******************************************************************/

/******************************************************************
***																***
***		Name:	Open BonDriver  								***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

BOOL OpenBonDriver( void )
{
								/*************************************/
								/**	Find BonDriver to load  		**/
								/*************************************/
	char szMyPath[_MAX_PATH] ;
	GetModuleFileName( hThis, szMyPath, _MAX_PATH ) ;
	char szDrv[_MAX_DRIVE], szDir[_MAX_DIR] ;
	_splitpath_s( szMyPath, szDrv, sizeof(szDrv),
		szDir, sizeof(szDir), NULL, 0, NULL, 0 ) ;
	char szMask[_MAX_PATH] ;
	sprintf_s( szMask, sizeof(szMask), "%s%sBonDriver*.dll", szDrv, szDir ) ;
	WIN32_FIND_DATA FindFileData ;
	HANDLE hFind = FindFirstFile( szMask, &FindFileData) ;
	if (hFind == INVALID_HANDLE_VALUE) {
		DebugPrintF( "ERROR : %s not found\n", szMask ) ;
		return FALSE ;
	}
    FindClose( hFind ) ;
	DebugPrintF( "Find %s\n", FindFileData.cFileName ) ;
								/*************************************/
								/**	create event for 2nd thread 	**/
								/*************************************/
	if ( !( hEventThread = CreateEvent( NULL, FALSE, FALSE, NULL ) ) ) {
		DebugPrintF( "ERROR : Fail to CreateEvent()\n" ) ;
		return FALSE ;
	}
								/*************************************/
								/**	open BonDriver  				**/
								/*************************************/
	HMODULE hDll = LoadLibrary( FindFileData.cFileName ) ;
	if ( !hDll ) {
		DebugPrintF( "ERROR : Fail to load %s\n", FindFileData.cFileName ) ;
		return FALSE ;
	}
								/*************************************/
								/**	get method  					**/
								/*************************************/
	P_CreateBonDriver pfCreateBonDriver =
		(P_CreateBonDriver)GetProcAddress( hDll, IBON_DRIVER2_METHOD  ) ;
	if ( !pfCreateBonDriver ) {
		DebugPrintF( "ERROR : %s() method not found.\n", IBON_DRIVER2_METHOD ) ;
		FreeLibrary( hDll ) ;
		return FALSE ;
	}
								/*************************************/
								/**	create tuner instance   		**/
								/*************************************/
	IBonDriver2* pBon = (*pfCreateBonDriver)() ;
	if ( !pBon ) {
		DebugPrintF( "ERROR : Fail to create tuner instance\n" ) ;
		FreeLibrary( hDll ) ;
		return FALSE ;
	}
								/*************************************/
								/**	open tuner  					**/
								/*************************************/
	if ( !pBon->OpenTuner() ) {
		DebugPrintF( "ERROR : Fail to open tuner\n" ) ;
		pBon->Release() ;
		FreeLibrary( hDll ) ;
		return FALSE ;
	}
								/*************************************/
								/**	start 2nd thread				**/
								/*************************************/
	if ( !( hThread = CreateThread( NULL, 0, TsGetThread,
		NULL, 0, NULL ) ) ) {
		DebugPrintF( "ERROR : Fail to CreateThread()\n" ) ;
		pBon->CloseTuner() ;
		pBon->Release() ;
		FreeLibrary( hDll ) ;
		return FALSE ;
	}
								/*************************************/
								/**	success to open tuner   		**/
								/*************************************/
	hDllBon = hDll ;
	pInsBon = pBon ;
	return TRUE ;
}

/******************************************************************
***																***
***		Name:	CloseBonDriver  								***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

void CloseBonDriver( void )
{
								/*************************************/
								/**	stop 2nd thread 				**/
								/*************************************/
	if ( hThread ) {
		if ( hEventThread ) SetEvent( hEventThread ) ;
		WaitForSingleObject( hThread, INFINITE ) ;
		CloseHandle( hThread ) ;
		hThread = NULL ;
	}
								/*************************************/
								/**	close bon driver				**/
								/*************************************/
	if ( hDllBon ) {
		if ( pInsBon ) {
			pInsBon->CloseTuner() ;
			pInsBon->Release() ;
			pInsBon = NULL ;
		}
		FreeLibrary( hDllBon ) ;
		hDllBon = NULL ;
	}
								/*************************************/
								/**	close event for 2nd thread  	**/
								/*************************************/
	if( hEventThread ) {
		CloseHandle( hEventThread ) ;
		hEventThread = NULL;	
	}
}

/******************************************************************
***																***
***		Name:	Hook channel setting							***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

void HookChannelSetting(
	LPVOID lpInBuffer, DWORD nInBufferSize )
{
								/*************************************/
								/**	channel table for sateliate 	**/
								/*************************************/
	const static struct {
		DWORD dwCode ;
		DWORD dwSpace ;
		DWORD dwChannel ;
		char* pszChannel ;
	} asSat[] = {
		{ 0x08324010, 1,  0, "BS1/TS0 BS" },
		{ 0x08324011, 1,  1, "BS1/TS1 BS-i" },
		{ 0x08804030, 1,  2, "BS3/TS0 WOWOW" },
		{ 0x08804031, 1,  3, "BS3/TS1 BSWp" },
		{ 0x09644090, 1,  4, "BS9/TS0 BS11" },
		{ 0x09644091, 1,  5, "BS9/TS1 Star Channel HV" },
		{ 0x09644092, 1,  6, "BS9/TS2 TwellV" },
		{ 0x0A0040D0, 1,  7, "BS13/TS0 BSe" },
		{ 0x0A0040D1, 1,  8, "BS13/TS1 BStW" },
		{ 0x0A4C40F1, 1,  9, "BS15/TS1 NHK BS1/2" },
		{ 0x0A4C40F2, 1, 10, "BS15/TS2 NHK BS-hi" },
		{ 0x0C9A0000, 2,  0, "ND2 110CS #1" },
		{ 0x0CEA0000, 2,  1, "ND4 110CS #2" },
		{ 0x0D3A0000, 2,  2, "ND6 110CS #3" },
		{ 0x0D8A0000, 2,  3, "ND8 110CS #4" },
		{ 0x0DDA0000, 2,  4, "ND10 110CS #5" },
		{ 0x0E2A0000, 2,  5, "ND12 110CS #6" },
		{ 0x0E7A0000, 2,  6, "ND14 110CS #7" },
		{ 0x0ECA0000, 2,  7, "ND16 110CS #8" },
		{ 0x0F1A0000, 2,  8, "ND18 110CS #9" },
		{ 0x0F6A0000, 2,  9, "ND20 110CS #10" },
		{ 0x0FBA0000, 2, 10, "ND22 110CS #11" },
		{ 0x100A0000, 2, 11, "ND24 110CS #12" },
		{ 0x00000000, 0,  0, "sȃ`l" },
	} ;
								/*************************************/
								/**	work variables  				**/
								/*************************************/
	BYTE* pyIn = (BYTE*)lpInBuffer ;
	DWORD dwSpace, dwChannel ;
	BOOL bSetChannel = FALSE ;
	static union {
		BYTE ay[4] ;
		DWORD dw ;
	} uSat ;
								/*************************************/
								/**	Get channel of terrestrial  	**/
								/*************************************/
	if ( nInBufferSize == 43 && pyIn[0] == 0x40 && pyIn[1] == 0x03 &&
		pyIn[41] == 0xb2 && pyIn[42] == 0x08 ) {
		WORD wChCode = ( WORD(pyIn[39]) << 8 ) + pyIn[40] ;
		DWORD dwChan = ( wChCode - 0x0e7f ) / 0x002a + 13 ;
		DebugPrintF( "Terrestrial:ch%u\n", dwChan ) ;
		dwSpace = 0 ;
		/******** FOR UNIDEN/MASPRO TUNER(12button Remo-Con) ONLY *******/
		/******** FOR UNIDEN/MASPRO TUNER(12button Remo-Con) ONLY *******/
		dwChannel = ( dwChan - 13 ) % 12 ;
		/******** FOR UNIDEN/MASPRO TUNER(12button Remo-Con) ONLY *******/
		/******** FOR UNIDEN/MASPRO TUNER(12button Remo-Con) ONLY *******/
		bSetChannel = TRUE ;
	}
								/*************************************/
								/**	Get channel of satelite 		**/
								/*************************************/
	else if ( nInBufferSize == 41 && pyIn[0] == 0x40 && pyIn[1] == 0x03 ) {
		if ( pyIn[39] == 0x00 ) uSat.ay[3] = pyIn[40] ;
		else if ( pyIn[39] == 0x01 ) uSat.ay[2] = pyIn[40] ;
	}
	else if ( nInBufferSize == 38 && pyIn[0] == 0x40 &&
		pyIn[1] == 0x01 && pyIn[3] == 0x32 ) {
		if ( pyIn[4] == 0x8f ) uSat.ay[1] = pyIn[2] ;
		else if ( pyIn[4] == 0x90 ) {
			uSat.ay[0] = pyIn[2] ;
			int i ;
			for ( i = 0 ; asSat[i].dwCode ; i++ ) {
				if ( asSat[i].dwCode == uSat.dw ) break ;
			}
			DebugPrintF( "Satelite:%s\n", asSat[i].pszChannel ) ;
			if ( asSat[i].dwCode ) {
				dwSpace = asSat[i].dwSpace ;
				dwChannel = asSat[i].dwChannel ;
				bSetChannel = TRUE ;
			}
		}
	}
								/*************************************/
								/**	set channel 					**/
								/*************************************/
	if ( bSetChannel ) {
		BonSetChannel( dwSpace, dwChannel ) ;
	}
}

/******************************************************************
***																***
***		Name:	Set channel 									***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

void BonSetChannel( DWORD dwSpace, DWORD dwChannel )
{
	DebugPrintF( "Space=%d/Channel=%d\n", dwSpace, dwChannel ) ;
	if ( pInsBon ) pInsBon->SetChannel( dwSpace, dwChannel ) ;
}

/******************************************************************
***																***
***		Name:	Purge TS data   								***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

void PurgeTsData( void )
{
	if ( pInsBon ) pInsBon->PurgeTsStream() ;
}

/******************************************************************
***																***
***		Name:	2nd thread to get TS data from BonDriver		***
***																***
*************************<< Function >>****************************
***																***
***																***
***																***
******************************************************************/

DWORD WINAPI TsGetThread( LPVOID lpParameter )
{
								/*************************************/
								/**	raise priority of this thread   **/
								/*************************************/
	SetThreadPriority( hThread, THREAD_PRIORITY_TIME_CRITICAL ) ;
	DWORD dTotal = 0 ;
	DWORD dSize, dRemain = 0 ;
	while ( TRUE ) {
								/*************************************/
								/**	check if terminate thread   	**/
								/*************************************/
		if ( !dRemain ) {
			if ( WaitForSingleObject( hEventThread, 100 ) != WAIT_TIMEOUT ) break ;
			DebugPrintF( "Total=%u\n", dTotal ) ;
		}
								/*************************************/
								/**	get TS data from BonDriver  	**/
								/*************************************/
		dRemain = 0 ;
		BYTE* pData ;
		dSize = dRemain = 0 ;
		if ( !pInsBon->GetTsStream( &pData, &dSize, &dRemain ) || !dSize ) continue ;
		while ( dSize ) {
			dTotal += dSize ;
			dSize = 0 ; // withdraw data in this step
		}
	}
	return 0 ;
}

